"""Module to hold the mixin for the managers."""

from abc import ABCMeta, abstractmethod

from plutto.resource_handlers import resource_all, resource_create, resource_get
from plutto.utils import can_raise_http_error, get_resource_class, pluralize


class ManagerMixin(metaclass=ABCMeta):  # pylint: disable=no-self-use
    """Class to hold the mixin for the managers."""

    def __init__(self, path, client):
        self._path = path
        self._client = client
        self._handlers = {
            "update": self.post_update_handler,
            "delete": self.post_delete_handler,
        }

    def __getattr__(self, attr):
        if attr not in self.__class__.methods:
            raise AttributeError(
                f"{self.__class__.__name__} has no attribute '{attr.lstrip('_')}'"
            )
        return getattr(self, f"_{attr}")

    @property
    @abstractmethod
    def resource(self):
        """
        This abstract property must be instanced as a class attribute
        when subclassing this mixin. It represents the name of the resource
        using snake_case.
        """

    @property
    @abstractmethod
    def methods(self):
        """
        This abstract property must be instanced as a class attribute
        when subclassing this mixin. It represents the methods that can be
        accessed using the manager. Must be an array with at leat one of:
        ['all', 'get', 'create', 'update', 'delete'].
        """

    @can_raise_http_error
    def _all(self, **kwargs):
        """
        Method to fetch all objects.

        :kwargs: can be used to filter the results using the API parameters.
        """
        klass = get_resource_class(self.__class__.resource)
        objects = resource_all(
            client=self._client,
            path=self._path,
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=pluralize(self.__class__.resource),
            params=kwargs,
        )
        return self.post_all_handler(objects, **kwargs)

    @can_raise_http_error
    def _get(self, identifier, **kwargs):
        """
        Return an instance of the resource being handled by the manager,
        indentified by :identifier:.
        """
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_get(
            client=self._client,
            path=self._path,
            id_=identifier,
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )
        return self.post_get_handler(object_, identifier, **kwargs)

    @can_raise_http_error
    def _create(self, **kwargs):
        """
        Create an instance of the resource being handled by the manager.
        Data is passed using :kwargs:, as specified by the API.
        """
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_create(
            client=self._client,
            path=self._path,
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )
        return self.post_create_handler(object_, **kwargs)

    @can_raise_http_error
    def _update(self, unique_identifier, **kwargs):
        """
        Update an instance of the resource being handled by the manager,
        identified by :identifier:. Data is passed using :kwargs:, as
        specified by the API.
        """
        object_ = self._get(unique_identifier)
        return object_.update(self.__class__.resource, **kwargs)

    @can_raise_http_error
    def _delete(self, identifier, **kwargs):
        """
        Delete an instance of the resource being handled by the manager,
        identified by :identifier:.
        """
        object_ = self._get(identifier)
        return object_.delete(**kwargs)

    def post_all_handler(self, objects, **kwargs):
        """
        Hook that runs after the :all: method. Receives the objects fetched
        and **must** return them (either modified or as they came).
        """
        return objects

    def post_get_handler(self, object_, identifier, **kwargs):
        """
        Hook that runs after the :get: method. Receives the object fetched
        with its identifier and **must** return the object (either modified
        or as it came).
        """
        return object_

    def post_create_handler(self, object_, **kwargs):
        """
        Hook that runs after the :create: method. Receives the object fetched
        and **must** return the it (either modified or as it came).
        """
        return object_

    def post_update_handler(self, object_, unique_identifier, **kwargs):
        """
        Hook that runs after the :update: method. Receives the object fetched
        with its identifier and **must** return the object (either modified
        or as it came).
        """
        return object_

    def post_delete_handler(self, identifier, **kwargs):
        """
        Hook that runs after the :create: method. Receives the identifier
        and **must** return it (either modified or as it came).
        """
        return identifier



"""Module for the methods that handle the resources."""

from plutto.utils import objetize, objetize_generator


def resource_all(client, path, klass, handlers, methods, resource, params):
    """Fetch all the instances of a resource."""
    lazy = params.pop("lazy", True)
    data = client.request(path, paginated=True, params=params, resource=resource)

    if lazy:
        return objetize_generator(
            data,
            klass,
            client,
            handlers=handlers,
            methods=methods,
            path=path,
        )

    return [
        objetize(
            klass,
            client,
            element,
            handlers=handlers,
            methods=methods,
            path=path,
        )
        for element in data
    ]


def resource_get(client, path, id_, klass, handlers, methods, resource, params):
    """Fetch a specific instance of a resource."""
    data = client.request(f"{path}/{id_}", method="get", params=params)[resource]
    return objetize(
        klass,
        client,
        data,
        handlers=handlers,
        methods=methods,
        path=path,
    )


def resource_create(client, path, klass, handlers, methods, resource, params):
    """Create a new instance of a resource."""
    data = client.request(path, method="post", json=params)[resource]
    return objetize(
        klass,
        client,
        data,
        handlers=handlers,
        methods=methods,
        path=path,
    )


def resource_update(client, path, id_, klass, handlers, methods, resource, params):
    """Update a specific instance of a resource."""
    data = client.request(f"{path}/{id_}", method="patch", json=params)[resource]
    return objetize(
        klass,
        client,
        data,
        handlers=handlers,
        methods=methods,
        path=path,
    )


def resource_delete(client, path, id_, params):
    """Delete a specific instance of a resource."""
    return client.request(f"{path}/{id_}", method="delete", params=params)


def resource_permission(client, path, id_, permission_name, klass, resource, params):
    """Fetch the permissions of a specific instance of a resource."""
    data = client.request(
        f"{path}/{id_}/has_permission/{permission_name}", method="get", params=params
    )[resource]

    return objetize(
        klass,
        client,
        data,
        path=path,
    )


def resource_patch(
    client, path, id_, action, klass, handlers, methods, resource, params
):
    """Patch to a specific endpoint of a resource"""
    data = client.request(f"{path}/{id_}/{action}", method="patch", json=params)[
        resource
    ]
    return objetize(
        klass,
        client,
        data,
        handlers=handlers,
        methods=methods,
        path=path,
    )



"""Module to hold the subscriptions manager."""

from plutto.mixins import ManagerMixin
from plutto.resource_handlers import resource_patch
from plutto.utils import can_raise_http_error, get_resource_class


class SubscriptionsManager(ManagerMixin):
    """Class to hold the subscriptions manager."""

    resource = "subscription"
    methods = ["get", "create", "end"]

    @can_raise_http_error
    def end(self, unique_identifier, **kwargs):
        """Ends the subscription which has unique identifier as id"""
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_patch(
            client=self._client,
            path=self._path,
            id_=unique_identifier,
            action="end_subscription",
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )

        return object_

    @can_raise_http_error
    def add_pricings(self, unique_identifier, **kwargs):
        """Adds pricings to the subscription with the unique identifier"""
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_patch(
            client=self._client,
            path=self._path,
            id_=unique_identifier,
            action="add_pricings",
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )

        return object_

    @can_raise_http_error
    def remove_pricings(self, unique_identifier, **kwargs):
        """Removes pricings to the subscription with the unique identifier"""
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_patch(
            client=self._client,
            path=self._path,
            id_=unique_identifier,
            action="remove_pricings",
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )

        return object_



import datetime
from importlib import import_module

import httpx

from plutto.errors import PluttoError


def singularize(string):
    """Remove the last 's' from a string if exists."""
    return string.rstrip("s")


def pluralize(string):
    """Add an 's' to a string if it doesn't already end with 's'."""
    if string.endswith("s"):
        return string
    return string + "s"


def snake_to_pascal(snake_string):
    """Convert a snake_case string to PascalCase."""
    return "".join(word.title() for word in snake_string.split("_"))


def get_resource_class(snake_resource_name, value={}):
    """
    Get the class that corresponds to a resource using its
    name (in snake case) and it's value.
    """
    if isinstance(value, dict):
        module = import_module("plutto.resources")
        try:
            return getattr(module, snake_to_pascal(snake_resource_name))
        except AttributeError:
            return getattr(module, "GenericPluttoResource")
    return type(value)


def get_error_class(function):
    """
    Given an error name (in snake case), return the appropiate
    error class.
    """
    module = import_module("plutto.errors")
    return getattr(module, snake_to_pascal(function), PluttoError)


def can_raise_http_error(function):
    """
    Decorator that catches HTTPError exceptions and raises custom
    Plutto errors instead.
    """

    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except httpx.HTTPError as e:
            error_data = e.response.json()
            error = get_error_class(error_data["error"]["type"])
            raise error(error_data["error"]) from None

    return wrapper


def serialize(object_):
    """Serialize an object."""
    if callable(getattr(object_, "serialize", None)):
        return object_.serialize()
    if isinstance(object_, datetime.datetime):
        return object_.isoformat()
    return object_


def objetize(klass, client, data, handlers={}, methods=[], path=None):
    """Transform the :data: object into an object with class :klass:"""
    if data is None:
        return None
    if klass in [str, int, dict, float, bool]:
        return klass(data)
    return klass(client, handlers, methods, path, **data)


def objetize_generator(generator, klass, client, handlers={}, methods=[], path=None):
    """
    Transform a generator of dictionaries into a generator of
    objects with class :klass:.
    """
    for element in generator:
        yield objetize(klass, client, element, handlers, methods, path)



"""Module to hold the invoices manager."""

from plutto.mixins import ManagerMixin
from plutto.resource_handlers import resource_patch
from plutto.utils import can_raise_http_error, get_resource_class


class InvoicesManager(ManagerMixin):
    """Class to hold the invoices manager."""

    resource = "invoice"
    methods = ["all", "get", "mark_as"]

    @can_raise_http_error
    def mark_as(self, unique_identifier, **kwargs):
        """Mark an invoice"""
        klass = get_resource_class(self.__class__.resource)
        object_ = resource_patch(
            client=self._client,
            path=self._path,
            id_=unique_identifier,
            action="mark_as",
            klass=klass,
            handlers=self._handlers,
            methods=self.__class__.methods,
            resource=self.__class__.resource,
            params=kwargs,
        )

        return object_



"""Module to hold the Plutto custom errors."""
from .constants import GENERAL_DOC_URL


class PluttoError(Exception):
    """Represents the base custom error."""

    def __init__(self, error_data, doc_url=GENERAL_DOC_URL):
        error_type = error_data.get("type")
        error_code = error_data.get("code")
        error_message = error_data.get("message")
        error_param = error_data.get("param")
        message = error_type
        message += f": {error_code}" if error_code else ""
        message += f" ({error_param})" if error_param else ""
        message += f"\n{error_message}"
        message += f"\nCheck the docs for more info: {doc_url}"

        super().__init__(message)


class InvalidRequestError(PluttoError):
    """Represents the invalid request error."""


class BadRequestError(PluttoError):
    """Represent the bad request error."""


class UnprocessableEntityError(PluttoError):
    """Represents the unprocessable entity error."""


class UnauthorizedError(PluttoError):
    """Represents the unauthorized entity error."""


class NotFoundError(PluttoError):
    """Represents the not found error."""


class InternalServerError(PluttoError):
    """Represents the internal server error."""


class AuthenticationError(PluttoError):
    """Represents the authentication error."""



"""Module to hold the customers manager."""

from plutto.mixins import ManagerMixin
from plutto.resource_handlers import resource_permission
from plutto.utils import can_raise_http_error, get_resource_class


class CustomersManager(ManagerMixin):
    """Class to hold the customers manager."""

    resource = "customer"
    methods = ["all", "get", "create", "update", "delete", "permission"]

    @can_raise_http_error
    def permission(self, unique_identifier, permission_name, **kwargs):
        """Check if the user has the permission with permission_name"""
        resource = "customer_permission"
        klass = get_resource_class(resource)
        object_ = resource_permission(
            client=self._client,
            path=self._path,
            id_=unique_identifier,
            permission_name=permission_name,
            klass=klass,
            resource=resource,
            params=kwargs,
        )

        return object_



"""
Module to house the Client object of the Plutto Python SDK.
"""

from json.decoder import JSONDecodeError

import httpx

from plutto.paginator import paginate


class Client:
    """Encapsulates the client behaviour and methods."""

    def __init__(self, base_url, api_key, user_agent, params={}):
        """Initializes the client object."""
        self.base_url = base_url
        self.api_key = api_key
        self.user_agent = user_agent
        self.params = params
        self.__client = None

    @property
    def _client(self):
        """Gets the client object."""
        if self.__client is None:
            self.__client = httpx.Client(
                base_url=self.base_url,
                headers=self.headers,
                params=self.params,
            )
        return self.__client

    @property
    def headers(self):
        """Gets the headers for the client."""
        return {
            "User-Agent": self.user_agent,
            "Authorization": f"Bearer {self.api_key}",
        }

    def request(
        self, path, paginated=False, method="get", params=None, json=None, resource=None
    ):
        """
        Uses the internal httpx client to make a simple request.
        """
        if paginated:
            return paginate(self._client, path, resource, params=params)

        response = self._client.request(method, path, params=params, json=json)
        response.raise_for_status()

        try:
            return response.json()
        except JSONDecodeError:
            return {}



"""Module to hold the products manager."""

from plutto.mixins import ManagerMixin


class ProductsManager(ManagerMixin):
    """Class to hold the products manager."""

    resource = "product"
    methods = ["all"]



"""Module to hold the constans of the SDK."""

API_BASE_URL = "https://app.getplutto.com/api"
API_VERSION = "v1"
GENERAL_DOC_URL = "https://docs.getplutto.com/"
LINK_HEADER_PATTERN = r'<(?P<url>.*)>;\s*rel="(?P<rel>.*)"'



"""Init file for the managers module of the SDK."""

from .customers_manager import CustomersManager
from .invoices_manager import InvoicesManager
from .meter_events_manager import MeterEventsManager
from .permission_groups_manager import PermissionGroupsManager
from .products_manager import ProductsManager
from .subscriptions_manager import SubscriptionsManager



import re
from functools import reduce

from plutto.constants import LINK_HEADER_PATTERN


def paginate(client, path, resource, params):
    """
    Fetch a paginated resource and return a generator with all of
    its instances.
    """
    response = request(client, path, params=params)
    elements = response["elements"][resource]
    for element in elements:
        yield element
    while response.get("next"):
        response = request(client, response.get("next"))
        elements = response["elements"][resource]
        for element in elements:
            yield element


def request(client, path, params={}):
    """
    Fetch a page of a resource and return its elements and the next
    page of the resource.
    """
    response = client.request("get", path, params=params)
    response.raise_for_status()
    headers = parse_link_headers(response.headers.get("link"))
    next_ = headers and headers.get("next")
    elements = response.json()

    return {
        "next": next_,
        "elements": elements,
    }


def parse_link_headers(link_header):
    """
    Receive the 'link' header and return a dictionary with
    every key: value instance present on the header.
    """
    if link_header is None:
        return None
    return reduce(parse_link, link_header.split(","), {})


def parse_link(dictionary, link):
    """
    Receive a dictionary with already parsed key: values from the
    'link' header along with another link and return the original
    dictionary with a new entry corresponding to the link received.
    """
    matches = re.match(LINK_HEADER_PATTERN, link.strip()).groupdict()
    dictionary[matches["rel"]] = matches["url"]
    return {**dictionary, matches["rel"]: matches["url"]}



"""Module to hold the meter events manager."""

from plutto.mixins import ManagerMixin


class MeterEventsManager(ManagerMixin):
    """Class to hold the meter events manager."""

    resource = "meter_event"
    methods = ["create"]



"""Module to hold the permission groups manager."""

from plutto.mixins import ManagerMixin


class PermissionGroupsManager(ManagerMixin):
    """Class to hold the permission groups manager."""

    resource = "permission_group"
    methods = ["all"]



"""Init file for the resources module of the SDK."""

from .customer import Customer
from .customer_permission import CustomerPermission
from .generic_plutto_resource import GenericPluttoResource
from .invoice import Invoice
from .meter_event import MeterEventResource
from .permission_group import PermissionGroup
from .product import Product
from .subscription import Subscription



from abc import ABCMeta

from plutto.resource_handlers import resource_delete, resource_update
from plutto.utils import (
    can_raise_http_error,
    get_resource_class,
    objetize,
    serialize,
    singularize,
)


class ResourceMixin(metaclass=ABCMeta):

    """Represents the mixin for the resources."""

    mappings = {}
    resource_identifier = "id"

    def __init__(self, client, handlers, methods, path, **kwargs):
        self._client = client
        self._handlers = handlers
        self._methods = methods
        self._path = path
        self._attributes = []
        for key, value in kwargs.items():
            try:
                resource = self.__class__.mappings.get(key, key)
                if isinstance(value, list):
                    resource = singularize(resource)
                    element = {} if not value else value[0]
                    klass = get_resource_class(resource, value=element)
                    setattr(self, key, [objetize(klass, client, x) for x in value])
                else:
                    klass = get_resource_class(resource, value=value)
                    setattr(self, key, objetize(klass, client, value))
                self._attributes.append(key)
            except NameError:  # pragma: no cover
                pass

    def __getattr__(self, attr):
        if attr not in self._methods:
            raise AttributeError(
                f"{self.__class__.__name__} has no attribute {attr.lstrip('_')}"
            )
        return getattr(self, f"_{attr}")

    def serialize(self):
        """Serialize the resource"""
        serialized = {}
        for key in self._attributes:
            element = (
                [serialize(x) for x in self.__dict__[key]]
                if isinstance(self.__dict__[key], list)
                else serialize(self.__dict__[key])
            )
            serialized = {**serialized, key: element}
        return serialized

    @can_raise_http_error
    def _update(self, resource, **kwargs):
        """Updates the resource."""
        id_ = getattr(self, self.__class__.resource_identifier)
        object_ = resource_update(
            client=self._client,
            path=self._path,
            id_=id_,
            klass=self.__class__,
            handlers=self._handlers,
            methods=self._methods,
            resource=resource,
            params=kwargs,
        )
        object_ = self._handlers.get("update")(object_, id_, **kwargs)
        self.__dict__.update(object_.__dict__)
        return self

    @can_raise_http_error
    def _delete(self, **kwargs):
        """Deletes the resource."""
        identifier = getattr(self, self.__class__.resource_identifier)
        resource_delete(
            client=self._client,
            path=self._path,
            id_=self.id,
            params=kwargs,
        )
        return self._handlers.get("delete")(identifier, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plutto', 'plutto.managers', 'plutto.mixins', 'plutto.resources']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.16,<1.0']

setup_kwargs = {
    'name': 'plutto',
    'version': '0.2.0',
    'description': 'The official Python client for Plutto API',
    'long_description': '<h1 align="center">Plutto Python Client</h1>\n\n<p align="center">\n  <a href="https://pypi.org/project/plutto" target="_blank">\n      <img src="https://img.shields.io/pypi/v/plutto?label=version&logo=python&logoColor=%23fff&color=306998" alt="PyPI - Version">\n  </a>\n\n  <a href="https://github.com/plutto-labs/plutto-python/actions?query=workflow%3Atests" target="_blank">\n      <img src="https://img.shields.io/github/workflow/status/plutto-labs/plutto-python/tests?label=tests&logo=python&logoColor=%23fff" alt="Tests">\n  </a>\n\n  <a href="https://codecov.io/gh/plutto-labs/plutto-python" target="_blank">\n      <img src="https://img.shields.io/codecov/c/gh/plutto-labs/plutto-python?label=coverage&logo=codecov&logoColor=ffffff" alt="Coverage">\n  </a>\n\n  <a href="https://github.com/plutto-labs/plutto-python/actions?query=workflow%3Alinters" target="_blank">\n      <img src="https://img.shields.io/github/workflow/status/plutto-labs/plutto-python/linters?label=linters&logo=github" alt="Linters">\n  </a>\n</p>\n\nThis library will help you easily integrate Plutto API to your software, making your developer life a little bit more enjoyable.\n\n\n---\n\n## Installation\nInstall using pip\n\n```bash\n$ pip3 install plutto\n```\n*Note:* This SDK requires [Python 3.7+](https://www.python.org/downloads/release/python-370/)\n\n## Usage\nWith this SDK we want to provide a wrapper to the [Plutto API](https://docs.getplutto.com/reference) with a very intuitive way of use. All the methods were implemented as in the API documentation, we strongly recommend to read it before using this SDK\n\n### Quickstart\nFirst of all, you will need a [Plutto] account. After creating it, you can get your API key, which will let you to use the `Plutto` object. Then, you\'re ready to use this awesome SDK!\n\n```python\nfrom plutto import Plutto\n\nclient = Plutto("your_api_key")\n```\n\n### Managers\nTo manage the resources retrieved by the SDK we use managers. They are python objects that let you with any object inside Plutto API.. All the existing **managers** are inside the `Plutto` object. These are:\n- `customers`\n- `invoices`\n- `meter_events`\n- `permission_groups`\n- `products`\n- `subscriptions`\n\n#### `all`\n_Note_: this method is only available in `customers`, `invoices`, `permission_groups` and `products` managers\n\n```python\ncustomers = client.customers.all()\n```\n\nThis method returns a **a generator** with all the instances of the customers resource. But, what if the API can recive more params? `kwargs` to the rescue! This way you can pass params like `q[status_eq]` and `q[customer_eq]` to filter the `Invoices`. If you want to get `invoices` with an specific _status_ and _customer_, you need to pass those params to the request\n```python\nparams = {\n    "q[status_eq]": "paid",\n    "q[customer_eq]": "customer_id"\n}\ninvoices = client.invoices.all(**params)\n```\n\nAlso, if you pass the `lazy=False` parameter, this will force the method of the SDK to return a list of the instances of the resource, instead of the generetors of them. **Disclaimer**: This could take very long if you have a lot of instances to be retrieved.\n\n```python3\ncustomers = client.customers.all(lazy=False)\nisinstance(customers, list) # True\n```\n\n#### `get`\n_Note_: this method is only available in `customers` and `invoices` managers\n\nThis method returns an instance of a resource using it\'s identifier to find it\n\n```python3\ncustomer = client.customers.get("customer_id")\nisinstance(customer, Customer) # True\n```\n\n#### `create`\n_Note_: this method is only available in `customers`, `meter_events` and `subscriptions` managers\n\nThis method creates and returns a new instance of the resource. The attributes of the resource to be created must be passed as `kwargs`. This parameters are specified in the API documentation of the correspondant resource\n\n```python3\npayload = {\n    "identifier": "your-id_12885305",\n    "email": "donald@getplutto.com",\n    "name": "Donald",\n    "billing_information": {\n        "city": "Santiago",\n        "country_iso_code": "CL",\n        "state": "Metropolitana",\n        "address": "Av. Las Condes",\n        "zip": "12345",\n        "tax_id": "73245432-1",\n        "legal_name": "Plutto Inc",\n        "activity": "Software Development",\n        "phone": "+56912345678"\n    }\n}\n\ncustomer = client.customer.create(**payload)\n```\n\n#### `update`\n_Note_: this method is only available in `customers` manager\n\n```python3\ncustomer = client.cuestomers.update(\n    "user_id",\n    email="goofy@getplutto.com",\n    name="Goofy"\n)\n```\nThis is an example of how can be used the update method. The first parameter corresponds to the id of the customer, this way the manager can find the existing resource. Then, the attributes that want to be modified are passed as `kwargs`, this ones are specified in the API in the correspondant resource update method.\nThis manager method is making two calls to the Plutto API, the first, to get the resource, and the second to update it. Therefore, if you only want to make one API call and already got the reource python object, you can call update directly on the object\n\n```python3\n# Get the object\nexample_customer = client.customers.get("customer_id")\n\n# Update the customer\nexample_customer.update(\n    email="goofy@getplutto.com",\n    name="Goofy"\n)\n```\n\nThis way, you can call `update` on the objects you are already working with, evitating to make an innescesary API call and saving some words\n\n\n#### `delete`\n_Note_: this method is only available in `customers` manager\n\n```python3\ndeleted_customer_id = client.customers.delete("customer_id")\n```\nThis method deletes an existing instance of a resource by it\'s identifier, and returns it. As in the `update` method, you can call `delete` on an resource object, for the same reasons explained in the previous method\n\n```python3\n# Get the resource\ncustomer = client.customers.get("customer_id")\n\n# Delete de resource\ndeleted_customer_id = customer.delete()\n```\n\n\n### How to use this SDK\nThe way to use is very similar to the API. For all the methods you need the `Plutto` object\n\n#### `Plutto` object\nInstantiate the object using your secret API key\n```python3\nfrom plutto import Plutto\n\nclient = Plutto("secret_api_key")\n```\nThis client will give you access to all the managers that are available in this SDK. That means you can work with any [manager](#managers) you want from this object\n\n#### `customers` manager\n\n_Available methods_: `all`, `get`, `create`, `update` and `delete`\n\nFrom the `Plutto` object you can manage your `customers` easily. You can get all the customers that have been created\n\n```python3\ncustomers = client.customers.all()\nfor customer in customers:\n    print(customer.name)\n```\n\nAlso, if you have the id, you can get a singular customer\n\n```python3\ncustomer = client.customers.get("id_of_the_customer")\n```\n\nNeed a new customer? Creating it is very intuitive, just pass the [parameters specified in the docs](https://docs.getplutto.com/reference/post_customers) as kwargs\n\n```python3\npayload = {\n    "email": "donald@getplutto.com",\n    "name": "donald",\n    "billing_information": {\n        "city": "Santiago",\n        "country_iso_code": "CL",\n        "legal_name": "Plutto",\n        "activity": "Software Development"\n    }\n}\n\nnew_customer = client.customers.create(**payload)\n```\n\nIf any customers need to be updated, you can do it with it\'s id and pass the params you want to update as kwargs\n\n```python3\nupdate_params = {\n    "name": "Goofy",\n    "email": "goofy@getplutto.com"\n}\n\nupdated_customer = client.customers.update("update_customer_id", **update_params)\n\n# It can also be done this way\nupdate_params = {\n    "name": "Goofy",\n    "email": "goofy@getplutto.com"\n}\n\ncustomer = client.customers.get("update_customer_id")\nupdated_customer = customer.update(**update_params)\n```\n\nDelete a customer can be done by passing it\'s id to the `delete` method\n\n```python3\ndeleted_customer_id = client.customers.delete("delete_customer_id")\n\n# It can also be done this way\ncustomer = client.customers.get("delete_customer_id")\ndeleted_customer_id = customer.delete()\n```\n\n\n#### `invoices` manager\n\n_Available methods_: `all`, `get` and `mark_as`\n\nTo get all the existing invoices you can do the following\n\n```python3\ninvoices = client.invoices.all()\n\nfor invoice in invoices:\n    print(invoice.id)\n\n# If filtering by params is needed\nparams = {\n    "q[status_eq]": "paid",\n    "q[customer_eq]": "customer_id"\n}\ninvoices = client.invoices.all(**params)\n```\n\nTo get a specific invoice, you can use it\'s id\n\n```python3\ninovice = client.invoices.get("invoice_id")\n```\n\nIt\'s possible to change an invoice status through the manager. The possible status are [specified in the docs](https://docs.getplutto.com/reference/patch_invoices-id-mark-as)\n\n```python3\ninvoice = client.invoices.mark_as("invoice_id", status="paid")\n```\n\n#### `meter_events` manager\n_Available methods_: Only `create`\n\nIt\'s possible to create a meter event this way\n\n```python3\npayload = {\n    "customer_id": "example_customer_id",\n    "meter_id": "example_meter_id",\n    "amout": 100,\n    "action": "set"\n}\nmeter_event = client.meter_events.create(**payload)\n```\n\n\n#### `permission_groups` manager\n_Available methods_: Only `all`\n\nYou can get all the permission groups as follows\n\n```python3\npermission_groups = client.permission_groups.all()\nfor permission_group in permission_groups:\n    print(permission_group.name)\n```\n\n\n#### `products` manager\n_Available methods_: Only `all`\n\nYou can get all the products as follows\n\n```python3\nproducts = client.products.all()\nfor product in products:\n    print(product.name)\n```\n\n\n#### `subscriptions` manager\n\nTo create a new subscription through the manager you can do the following. The attributes must be passed as kwargs. Required and optional ones are [specified in the docs](https://docs.getplutto.com/reference/post_subscriptions)\n\n```python3\npayload = {\n    "customer_id": "example_customer_id",\n    "pricing_ids": ["example_pricing_id_1", "example_pricing_id_2"]\n    "bills_at": "start",\n    "billing_period_duration": "P0Y1M0DT0H0M0S"\n}\nsubscription = client.subscriptions.create(**payload)\n```\n\nTo end a subscription you only need the id\n\n```python3\nsubscription = client.subscriptions.end("subscription_id")\n```\n\nAdding a product pricing to a subscription\n\n```python3\n# pricing_ids can have only one item, but it must be a list\npayload = {\n    "pricing_ids": ["pricing_id_1", "pricing_id_2", "pricing_id_3"]\n}\n\nsubscription = client.subscriptions.add_pricings("subscription_id", **payload)\n```\n\nTo remove a pricing, you can do the following\n\n```python3\n# pricing_ids can have only one item, but it must be a list\npayload = {\n    "pricings_id": ["pricing_id_1", "pricing_id_2", "pricing_id_3"]\n}\n\nsubscription = client.subscriptions.remove_pricings("subscription_id", **payload)\n```\n\n\n## Serialization\nAny resource retrieved by the SDK can be serialize, and it\'s super easy to do it. You just need to call the `serialize` method, and it\'s available in any resource\n\n```python3\ncustomer = client.customers.get("customer_id")\nserialized_customer = customer.serialize()\n```\n\n`serialized_customer` corresponds to a dictionary with only the attributes of the retrieved resource. It can be JSON-serialized\n\n\n## Testing\nAll the tests must be added in the `tests/` directory. To run the tests you nedd to execute the following command on the root path of the plutto library\n\n```bash\npytest .\n```\n\n\nEvery piece of code modified or added must be tested. The coverage always have to be increased or maintained, this will be checked in all PR\n\n## Publishing\n\nOn master/main branch...\n\n1. Change `VERSION` in `plutto/version.py`.\n2. Change `Unreleased` title to current version in `CHANGELOG.md`.\n3. Commit new release. For example: `Releasing v0.1.0`.\n4. Create tag. For example: `git tag v0.1.0`.\n5. Push tag. For example: `git push origin v0.1.0`.\n\n## Contributing\n\n1. Fork it\n2. Create your feature branch (`git checkout -b my-new-feature`)\n3. Commit your changes (`git commit -am \'Add some feature\'`)\n4. Push to the branch (`git push origin my-new-feature`)\n5. Create new Pull Request\n\n## Credits\n\nThank you [contributors](https://github.com/plutto-labs/plutto-python/graphs/contributors)!\n\nPlutto Ruby SDK is maintained by [Plutto](https://getplutto.com).\n\n\n## Acknowledgments\n\nThis SDK was strongly based on the [Fintoc python\'s SDK](https://github.com/fintoc-com/fintoc-python), designed by [Daniel Leal](https://github.com/daleal)\n## License\n\nPlutto Python SDK is © 2021 plutto, spa. It is free software and may be redistributed under the terms specified in the LICENSE file.',
    'author': 'Nicolas Massardo',
    'author_email': 'nmassardot@uc.cl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://getplutto.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Core module to house the Plutto object of the Plutto Python SDK.
"""

from plutto.client import Client
from plutto.constants import API_BASE_URL, API_VERSION
from plutto.managers import (
    CustomersManager,
    InvoicesManager,
    MeterEventsManager,
    PermissionGroupsManager,
    ProductsManager,
    SubscriptionsManager,
)
from plutto.version import __version__


class Plutto:

    """Encapsulates the core object's behaviour and methods."""

    def __init__(self, api_key):
        self._client = Client(
            base_url=f"{API_BASE_URL}/{API_VERSION}",
            api_key=api_key,
            user_agent=f"plutto-python/{__version__}",
        )

        self.__customers_manager = None
        self.__subscriptions_manager = None
        self.__meter_events_manager = None
        self.__invoices_manager = None
        self.__products_manager = None
        self.__permission_groups_manager = None

    @property
    def customers(self):
        """Proxies the customers manager."""
        if self.__customers_manager is None:
            self.__customers_manager = CustomersManager("/customers", self._client)

        return self.__customers_manager

    @customers.setter
    def customers(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")

    @property
    def subscriptions(self):
        """Proxies the subscriptions manager."""
        if self.__subscriptions_manager is None:
            self.__subscriptions_manager = SubscriptionsManager(
                "/subscriptions", self._client
            )

        return self.__subscriptions_manager

    @subscriptions.setter
    def subscriptions(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")

    @property
    def meter_events(self):
        """Proxies the meter events manager."""
        if self.__meter_events_manager is None:
            self.__meter_events_manager = MeterEventsManager(
                "/meter_events", self._client
            )

        return self.__meter_events_manager

    @meter_events.setter
    def meter_events(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")

    @property
    def invoices(self):
        """Proxies the invoices manager."""
        if self.__invoices_manager is None:
            self.__invoices_manager = InvoicesManager("/invoices", self._client)

        return self.__invoices_manager

    @invoices.setter
    def invoices(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")

    @property
    def products(self):
        """Proxies the products manager."""
        if self.__products_manager is None:
            self.__products_manager = ProductsManager("/products", self._client)

        return self.__products_manager

    @products.setter
    def products(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")

    @property
    def permission_groups(self):
        """Proxies the permission groups manager."""
        if self.__permission_groups_manager is None:
            self.__permission_groups_manager = PermissionGroupsManager(
                "/permission_groups", self._client
            )

        return self.__permission_groups_manager

    @permission_groups.setter
    def permission_groups(self, value):  # pylint: disable=no-self-use
        raise NameError("Attribute name corresponds to a manager")
